require("version.nut");
require("town.nut");

/* Import SuperLib for GameScript */
import("util.superlib", "SuperLib", 27);
Result <- SuperLib.Result;
Log <- SuperLib.Log;
Helper <- SuperLib.Helper;
ScoreList <- SuperLib.ScoreList;
Tile <- SuperLib.Tile;
Direction <- SuperLib.Direction;
Town <- SuperLib.Town;
Industry <- SuperLib.Industry;

class MainClass extends GSController 
{
	constructor()
	{
	}

	towns = null;
	last_date = 0;
	current_month = 0;
	current_year = 0;
}

function MainClass::Start()
{
	local start_tick = GSController.GetTick();
	GSGame.Pause();	
	this.Init();
	GSGame.Unpause();
	local setup_duration = GSController.GetTick() - start_tick;

	Log.Info("Game setup done", Log.LVL_INFO);
	Log.Info("Setup took " + setup_duration + " ticks", Log.LVL_INFO);
	Log.Info("", Log.LVL_INFO);

	Log.Info("Happy playing", Log.LVL_INFO);

	GSController.Sleep(1);

	while (true) {
		this.HandleEvents();
		this.ManageTowns2();

		this.Sleep(60);
	}
}

function MainClass::Init()
{
	GSGameSettings.SetValue("economy.town_growth_rate", 0);
	GSGameSettings.SetValue("economy.fund_buildings", 0);
	this.last_date = GSDate.GetCurrentDate();
	this.current_month = GSDate.GetMonth(this.last_date);
	this.current_year = GSDate.GetYear(this.last_date);
	
	// Clear all signs if this is a saved game.
	Helper.ClearAllSigns();

	// Setup Global Cargo Variables
		local IndustrySetting = GSController.GetSetting("Industry_Newgrf");
		
		switch(IndustrySetting){
		case(1): ::CargoList <- [1,0,2,5,10]; //Base industry
				 ::CargoMinPopDemand <- [0,500,1000,1500,2500]; 
				 ::Permille <- [120,250,45,250,13];
				 ::CargoDecay <- [0.1,0.5,0.5,0.05,0.02];
				 break;
		case(2): ::CargoList <- [0,2,11,10,5,29,25,28,18,24,1,31]; //FIRS
				 ::CargoMinPopDemand <- [0,0,1000,2000,3500,5500,8000,11000,15000,20000,25000,30000];
				 ::Permille <- [144,24,15,5,30,20,30,30,15,15,15,20];
				 ::CargoDecay <- [0.5,0.5,0.1,0.1,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05];
				 break;
		case(3): ::CargoList <- [0,2,3,20,5,25,19,18,9,1,24,10]; //ECS
				 ::CargoMinPopDemand <- [0,0,1000,2000,4000,6000,10000,15000,20000,25000,30000,35000];
				 ::Permille <- [144,24,15,15,25,15,15,15,25,15,5,10];
				 ::CargoDecay <- [0.5,0.5,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.02,0.02];
				 break;
		}
		
	// Create population tier limits
	::PopTierLimits <- [1000,3000,6000,10000,15000,21000,28000,36000,45000,55000,66000,78000,91000,105000,120000,136000,153000,171000,190000,210000,231000,253000,276000,300000,325000,351000,378000,406000,435000,465000,496000,528000,561000,595000,630000,666000,703000,741000,780000,820000,861000,903000,946000,990000];
	

	// Create the towns list
	Log.Info("Create town list ... (can take a while on large maps)", Log.LVL_INFO);
	this.towns = this.CreateTownList();

	::MinPeerCities <- max((sqrt(GSMap.GetMapSize())/128).tointeger(),3);
	
	local ModifierGrowth = (1/::MinPeerCities.tofloat());

	::ModifierList <- [0.0];
	
	for (local i=1;i<::MinPeerCities+1;++i)
	{
		::ModifierList.append(pow(0.5,pow(((i.tofloat()-1)/::MinPeerCities.tofloat()),2))*ModifierGrowth+::ModifierList[i.tofloat()-1]);
	}

	this.YearlyPopCheck();
	
	// Set initial time the towns will grow
	InitalizeManageTimers(this.towns);
	CheckCargo();
	
}




function MainClass::CheckCargo(){

local list = GSCargoList();

// Helper function to get a list of currently used cargo labels and their ID.
for(local i = 0; i < 32; ++i){
GSLog.Info(i + " = " + GSCargo.GetCargoLabel(i));
	}
}

// Non-class function
function InitalizeManageTimers(town_list)
{
	foreach (town in town_list) {
		town.check_timer = GSBase.RandRange(20) + 10;
		town.MonthlyManageTown();
	}
}

function MainClass::HandleEvents()
{
	if (GSEventController.IsEventWaiting())
	{
		local ev = GSEventController.GetNextEvent();

		if (ev == null)
			return;

		// check event type
		// handle event
	}
}

function MainClass::CreateTownList()
{
	local result = [];

	/* Make a squirrel array of all towns.
	 * For each town, an instance of GoalTown is created
	 * to store the data related to that town
	 */
	local all_towns = GSTownList();
	foreach (t, _ in all_towns)
	{
		result.append(GoalTown(t));
	}

	return result;
}

function MainClass::CheckForNewTowns()
{
	// Do our town-list have fewer towns than the game?
	local world_towns = GSTownList();
	if (this.towns.len() < world_towns.Count())
	{
		// Create a GSList with all towns that we have
		local have_towns = GSList();
		foreach (town in this.towns)
		{
			have_towns.AddItem(town.id, 0);
		}
		
		// Reduce the world list of towns with those that we already have
		world_towns.RemoveList(have_towns);

		// Add the remaining towns to our list
		// Also, keep a list of newly added towns
		local new_towns = [];
		foreach (town_id, _ in world_towns)
		{
			local town = GoalTown(town_id);
			this.towns.append(town);
			new_towns.append(town);
		}

		// Perform an initial manage of the new towns
		InitalizeManageTimers(new_towns);
	}
}

function MainClass::ManageTowns2()
{
	local date = GSDate.GetCurrentDate();
	local diff = date - this.last_date;
	this.last_date = date;
	local month = GSDate.GetMonth(date);
	local diff_month = current_month - month;
	local year = GSDate.GetYear(date);
	local diff_year = current_year - year;
	local tmp_rcvd = 0;


	if (diff == 0) return;
	// Run the daily expand town functions
	foreach (town in this.towns) 
	{
		town.ExpandTown();
	}
	
	// if month doesn't equal current month, execute MonthlyManageTown
	if ( diff_year == 0) 
	{
		this.YearlyPopCheck();
	}
	if ( diff_month == 0) return;
	GSGame.Pause();
	local month_tick = GSController.GetTick();
	GSLog.Info("Starting Monthly Updates");
	foreach (town in this.towns)
	{
		town.MonthlyManageTown();
	}
	
	current_month = month;
	local month_tick_duration = GSController.GetTick() - month_tick;
	GSLog.Info("Monthly Update took "+month_tick_duration+"ticks");
	GSGame.Unpause();
}

function MainClass::YearlyPopCheck()
{

	///////////////////////////////////////////////////////////////////////////////////////////
	// This function will loop through each town and assign it to a tier based on population //
	// for each tier we calculate the total number of towns in the same tier or higher       //
	///////////////////////////////////////////////////////////////////////////////////////////
	
	::TownTierCount <- [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
	::TownTierModifier <- [1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0];
	local TierLen = ::TownTierCount.len();
	local current_tier = TierLen - 1;
	local previous_tier = 0;
	local tier_temp = 0;
	local TierCumTemp = 0;
	
	foreach (town in this.towns)
	{
		town.YearlyManageTown();
	}
	
	TierCumTemp = ::TownTierCount[current_tier];
	for(local i = 1; i < TierLen; ++i)
	{
		previous_tier = current_tier
		current_tier = current_tier - 1;
		TierCumTemp += ::TownTierCount[current_tier];
			
		// If the number of cities at or above the current tier is
		// more than twice the minpeer cities, set modifier to the highest
		if(TierCumTemp > ::MinPeerCities*2)
		{
			::TownTierModifier[current_tier] = ::ModifierList[::MinPeerCities];
		}
		
		// Number of cities is above MinPeerCities, but below the max
		if(TierCumTemp > ::MinPeerCities && TierCumTemp <= ::MinPeerCities*2)
		{
			tier_temp = TierCumTemp - ::MinPeerCities;
			::TownTierModifier[current_tier] = ::ModifierList[tier_temp];
		}

		// Number of cities is equal to the target
		if(TierCumTemp == ::MinPeerCities)
		{
			::TownTierModifier[current_tier] = 1;
		}

		// Number of cities is below the target
		if(TierCumTemp < ::MinPeerCities)
		{
			tier_temp = ::MinPeerCities - TierCumTemp;
			::TownTierModifier[current_tier] = (1 + ::ModifierList[tier_temp]);
		}		
	}
}