
/*
 * For each town in the game, an instance of this class is created.
 * It holds the data related to a specific town.
 */
class GoalTown
{
	id = 0;
	rate = 0;
	check_timer = 0;
	TownLastGrowth = null;
	TownGrowthRate = null;
	TownStockPiles = [];
	TownSupplied = [];
	TownGoals = [];
	label_sign_id = -1;
	monitorcheck = false;
	signcheck = true;
	tier = 0;
	
	constructor(town_id)
	{
		TownLastGrowth = GSDate.GetCurrentDate();
		TownGrowthRate = 9999;
		local IndustrySetting = GSController.GetSetting("Industry_Newgrf");
		
		// Setup all the town specific arrays
		switch(IndustrySetting){
		case(1): TownStockPiles = [0,0,0,0,0];  // BASE
				 TownSupplied = [0,0,0,0,0];
				 TownGoals = [0,0,0,0,0];
				 break;
		case(2): TownStockPiles = [0,0,0,0,0,0,0,0,0,0,0,0]; // FIRS
				 TownSupplied = [0,0,0,0,0,0,0,0,0,0,0,0];
				 TownGoals = [0,0,0,0,0,0,0,0,0,0,0,0];
				 break;
		case(3): TownStockPiles = [0,0,0,0,0,0,0,0,0,0,0,0]; // ECS
				 TownSupplied = [0,0,0,0,0,0,0,0,0,0,0,0];
				 TownGoals = [0,0,0,0,0,0,0,0,0,0,0,0];
				 break;
		}
		
		this.id = town_id;
		this.rate = 1;
		GSTown.SetGrowthRate(this.id, 30000);
		local all_towns = GSTownList();
	}

	function MonthlyManageTown(); // <-- Main function to process town monthly

}

function GoalTown::MonthlyManageTown()
{
	local tempvar = ::CargoList.len();
	local sumgoals = 0;
	local goaldiff = 0;
	local goaldiffpercent = 0.0;
	local stocklimit = 0;
	local opsleft = 0;
	local paxcheck = 0;
	local curpop = GSTown.GetPopulation(this.id);
	

	
	// Difficulty factor:
	local d_percent = GSController.GetSetting("goal_scale_factor");
	local d_factor = d_percent > 0? d_percent / 100.0 : 1.0; // Use 1.0 if setting is missing


	for (local cid = GSCompany.COMPANY_FIRST; cid <= GSCompany.COMPANY_LAST; cid++)
	{
		if (GSCompany.ResolveCompanyID(cid) != GSCompany.COMPANY_INVALID)
		{
			paxcheck += GSCargoMonitor.GetTownPickupAmount(cid, ::CargoList[0],this.id,true);
		}
	}
	
	if(paxcheck == 0)
	{
		if(this.monitorcheck == true)
		{
			this.TownGrowthRate = 9999;
			this.StopMonitors();
			monitorcheck = false;
//			this.UpdateTownText();
		}
		return;
	}

	// Loops through each cargotype and company and sets the CargoMonitor for this town.
	for (local i = 0; i< tempvar; i++)
	{
		this.TownSupplied[i] = 0;  // Reset the supplied amount
		for (local cid = GSCompany.COMPANY_FIRST; cid <= GSCompany.COMPANY_LAST; cid++)
		{
			if (GSCompany.ResolveCompanyID(cid) != GSCompany.COMPANY_INVALID)
			{
					this.TownSupplied[i] += GSCargoMonitor.GetTownDeliveryAmount(cid, ::CargoList[i],this.id,true);
			}
		}
				
			if (this.TownGoals[i] > 0) // Only add supplied to stockpile if cargo is part of the town's goals.
			{
				this.TownStockPiles[i] += this.TownSupplied[i];
			}
		//////////////////////////////
		// Update TownGoals section //
		//////////////////////////////			
		
		this.TownGoals[i] = 0;
		if (curpop > ::CargoMinPopDemand[i])
		{
			if ( i < 2 )
				{
					this.TownGoals[i] = (GSTown.GetLastMonthProduction(this.id, ::CargoList[i]).tofloat() * 0.4 * d_factor).tointeger();
				}
				else
				{
					this.TownGoals[i] = max((((curpop - ::CargoMinPopDemand[i]).tofloat() / 1000) * ::Permille[i] * d_factor * ::TownTierModifier[this.tier]).tointeger(),1);
				}
		}

		///////////////////////////////////////////////////
		// This section calculates the sum of differences//
		// between supplied + stockpiled and the goal//////
		///////////////////////////////////////////////////		
		if ( this.TownGoals[i] != 0)
		{
			sumgoals += this.TownGoals[i];
		}
		if ( this.TownStockPiles[i] < this.TownGoals[i] )
		{
			goaldiff += (this.TownGoals[i] - this.TownStockPiles[i]);
		}
		
		if (this.TownGoals[i] > this.TownStockPiles[i])
		{
			this.TownStockPiles[i] = 0;
		}
		else
		{
			this.TownStockPiles[i] = ((this.TownStockPiles[i] - this.TownGoals[i]) * (1 - ::CargoDecay[i])).tointeger();
		}
		
		//stocklimit = this.TownGoals[i] * 10;
		if (this.TownStockPiles[i] > 300 && this.TownStockPiles[i] > this.TownGoals[i] * 10)
		{
			this.TownStockPiles[i] = this.TownGoals[i] * 10;
		}
	}
	
	// Convert the difference to a percent
	if ( sumgoals > 0 )
	{
		goaldiffpercent = (goaldiff.tofloat() / sumgoals.tofloat());
	}
	
	// Calculate the number of days till the next growth
	this.TownGrowthRate = 9999;
	if (goaldiffpercent < 0.2 && sumgoals > 0)
	{
		this.TownGrowthRate = (((1 / curpop.tofloat() ) * 10000)/(1-(goaldiffpercent*5))).tointeger();
	}

	// If the number of days between growths was less than one, reset it to one.
//	if ( this.TownGrowthRate < 1) { this.TownGrowthRate = 1; }
	

//	opsleft = GSController.GetOpsTillSuspend();
//	GSLog.Info("The number of ops remaining is: "+opsleft);
	monitorcheck = true;
	this.UpdateTownText();
	
}

function GoalTown::YearlyManageTown()
{
	local yearlypop = GSTown.GetPopulation(this.id);
	local testdone = true;
	local current_tier = 0;
	
	while(testdone)
	{
		if ( yearlypop < ::PopTierLimits[current_tier] )
		{
			::TownTierCount[current_tier] += 1;
			this.tier = current_tier;
			testdone = false;
			//GSLog.Info(GSTown.GetName(this.id)+" Is at tier "+this.tier);
		}
		else
		{
			current_tier += 1;
		}
	}
	
}

function GoalTown::ExpandTown()
{
	local today = GSDate.GetCurrentDate();
	local tmp_tg = today - this.TownGrowthRate + 1;
		
	if (tmp_tg >= this.TownLastGrowth)
	{
		GSTown.ExpandTown(this.id,1);
			this.TownLastGrowth = today;
			GSLog.Info(GSTown.GetName(this.id)+" has grown on "+today);
	}
			
}

function GoalTown::UpdateTownText()
{

	// Town text will look like below:
	// Label:stored  supplied  required
	// Coal:0/100  PAX:1/101  Mail:2/102  Gas:3/100  Goods:4/104
	local text = "";
	
	// Add a sign by the town to display the current growth
	local sign_text = "";
	sign_text = "Growth: "+this.TownGrowthRate+" days";
	
	if(this.label_sign_id == -1)
	{
		this.label_sign_id = GSSign.BuildSign(GSTown.GetLocation(this.id), sign_text);
	}
	else
	{
		if(GSSign.IsValidSign(this.label_sign_id))
		{
			GSSign.SetName(this.label_sign_id, sign_text);
		}
		else	
		{
			this.label_sign_id = GSSign.BuildSign(GSTown.GetLocation(this.id), sign_text);
		}
	}	
	
	if (this.signcheck == true)
	{
		for (local j = 0; j< ::CargoList.len(); j++)
		{
			text = text+GSCargo.GetCargoLabel(::CargoList[j])+":"+this.TownStockPiles[j]+"/"+this.TownSupplied[j]+"/"+this.TownGoals[j]+" ";
		}
		
		// Sets the towns text box
		GSTown.SetText(this.id, text);
		GSLog.Info(GSTown.GetName(this.id)+" text length is "+text.len());		
		// If the town text box gets too long, start sending text to signs instead
		
		// If the town text length is over 200, start using signs instead.
		if(text.len() > 200)
		{
			this.signcheck = false;
			GSTown.SetText(this.id,"");
		}

	}
	else
	{
		local tile_temp = GSTown.GetLocation(this.id);
		local x = GSMap.GetTileX(tile_temp);
		local y = GSMap.GetTileY(tile_temp)+1;
		for (local k = 0; k< ::CargoList.len(); k++)
			{
				text = GSCargo.GetCargoLabel(::CargoList[k])+":"+this.TownStockPiles[k]+"/"+this.TownSupplied[k]+"/"+this.TownGoals[k];
				Helper.SetSign(GSMap.GetTileIndex(x+k,y+k),text,true);
			}
	}
}

function GoalTown::StopMonitors()
{
	for (local i = 0; i< ::CargoList.len(); i++)
	{
		for (local cid = GSCompany.COMPANY_FIRST; cid <= GSCompany.COMPANY_LAST; cid++)
		{
			if (GSCompany.ResolveCompanyID(cid) != GSCompany.COMPANY_INVALID)
			{
				GSCargoMonitor.GetTownDeliveryAmount(cid, ::CargoList[i],this.id,false);
			}
		}
	}
}
